@@ -1,20 +1,20 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 import time
-from xml import sax
-from decimal import Decimal
 import datetime
+from xml import sax
 import logging
-import traceback
-import sys
 import re
-from sql import Table
-from itertools import izip
-from collections import defaultdict
 
-from .version import VERSION
-from .tools import safe_eval
+from collections import defaultdict
+from decimal import Decimal
+
+from . import __version__
+from .tools import grouped_slice
 from .transaction import Transaction
+from .pyson import PYSONEncoder, CONTEXT
+
+logger = logging.getLogger(__name__)
 
 CDATA_START = re.compile('^\s*\<\!\[cdata\[', re.IGNORECASE)
 CDATA_END = re.compile('\]\]\>\s*$', re.IGNORECASE)
@@ -44,18 +44,19 @@ class MenuitemTagHandler:
         self.xml_id = None
 
     def startElement(self, name, attributes):
-        cursor = Transaction().cursor
+        cursor = Transaction().connection.cursor()
 
         values = {}
 
         self.xml_id = attributes['id']
 
-        for attr in ('name', 'icon', 'sequence', 'parent', 'action', 'groups'):
-            if attributes.get(attr):
+        for attr in ('name', 'sequence', 'parent', 'action', 'groups'):
+            if attr in attributes:
                 values[attr] = attributes.get(attr)
+        values['icon'] = attributes.get('icon', 'tryton-folder')
 
         if attributes.get('active'):
-            values['active'] = bool(safe_eval(attributes['active']))
+            values['active'] = bool(eval(attributes['active']))
 
         if values.get('parent'):
             values['parent'] = self.mh.get_id(values['parent'])
@@ -65,14 +66,15 @@ class MenuitemTagHandler:
             action_id = self.mh.get_id(values['action'])
 
             # TODO maybe use a prefetch for this:
-            action = Table('ir_action')
-            report = Table('ir_action_report')
-            act_window = Table('ir_action_act_window')
-            wizard = Table('ir_action_wizard')
-            url = Table('ir_action_url')
-            act_window_view = Table('ir_action_act_window_view')
-            view = Table('ir_ui_view')
-            icon = Table('ir_ui_icon')
+            action = self.mh.pool.get('ir.action').__table__()
+            report = self.mh.pool.get('ir.action.report').__table__()
+            act_window = self.mh.pool.get('ir.action.act_window').__table__()
+            wizard = self.mh.pool.get('ir.action.wizard').__table__()
+            url = self.mh.pool.get('ir.action.url').__table__()
+            act_window_view = self.mh.pool.get(
+                'ir.action.act_window.view').__table__()
+            view = self.mh.pool.get('ir.ui.view').__table__()
+            icon = self.mh.pool.get('ir.ui.icon').__table__()
             cursor.execute(*action.join(
                     report, 'LEFT',
                     condition=action.id == report.action
@@ -88,9 +90,11 @@ class MenuitemTagHandler:
                     condition=view.id == act_window_view.view
                     ).join(icon, 'LEFT',
                     condition=action.icon == icon.id).select(
-                    action.name, action.type,
-                    view.type, view.field_childs,
-                    icon.name,
+                    action.name.as_('action_name'),
+                    action.type.as_('action_type'),
+                    view.type.as_('view_type'),
+                    view.field_childs.as_('field_childs'),
+                    icon.name.as_('icon_name'),
                     where=(report.id == action_id)
                     | (act_window.id == action_id)
                     | (wizard.id == action_id)
@@ -107,7 +111,7 @@ class MenuitemTagHandler:
             elif icon_name:
                 values['icon'] = icon_name
             elif action_type == 'ir.action.wizard':
-                values['icon'] = 'tryton-executable'
+                values['icon'] = 'tryton-launch'
             elif action_type == 'ir.action.report':
                 values['icon'] = 'tryton-print'
             elif action_type == 'ir.action.act_window':
@@ -117,15 +121,15 @@ class MenuitemTagHandler:
                     else:
                         values['icon'] = 'tryton-list'
                 elif view_type == 'form':
-                    values['icon'] = 'tryton-new'
+                    values['icon'] = 'tryton-form'
                 elif view_type == 'graph':
                     values['icon'] = 'tryton-graph'
                 elif view_type == 'calendar':
                     values['icon'] = 'tryton-calendar'
             elif action_type == 'ir.action.url':
-                values['icon'] = 'tryton-web-browser'
+                values['icon'] = 'tryton-public'
             else:
-                values['icon'] = 'tryton-new'
+                values['icon'] = None
 
         if values.get('groups'):
             raise Exception("Please use separate records for groups")
@@ -154,7 +158,7 @@ class MenuitemTagHandler:
             return None
 
     def current_state(self):
-        return "Tag menuitem with id: %s" % self.xml_id
+        return "Tag menuitem with id %s.%s" % (self.mh.module, self.xml_id)
 
 
 class RecordTagHandler:
@@ -198,10 +202,15 @@ class RecordTagHandler:
 
             field_name = attributes['name']
             field_type = attributes.get('type', '')
+            # Remind the current name and if we have to load (see characters)
+            self.current_field = field_name
+            depends = attributes.get('depends', '').split(',')
+            depends = {m.strip() for m in depends if m}
+            if not depends.issubset(self.mh.modules):
+                self.current_field = None
+                return
             # Create a new entry in the values
             self.values[field_name] = ""
-            # Remind the current name (see characters)
-            self.current_field = field_name
             # Put a flag to escape cdata tags
             if field_type == "xml":
                 self.cdata = "start"
@@ -210,12 +219,13 @@ class RecordTagHandler:
             search_attr = attributes.get('search', '')
             ref_attr = attributes.get('ref', '')
             eval_attr = attributes.get('eval', '')
+            pyson_attr = bool(int(attributes.get('pyson', '0')))
 
             if search_attr:
                 search_model = self.model._fields[field_name].model_name
                 SearchModel = self.mh.pool.get(search_model)
                 with Transaction().set_context(active_test=False):
-                    found, = SearchModel.search(safe_eval(search_attr))
+                    found, = SearchModel.search(eval(search_attr))
                     self.values[field_name] = found.id
 
             elif ref_attr:
@@ -224,10 +234,16 @@ class RecordTagHandler:
             elif eval_attr:
                 context = {}
                 context['time'] = time
-                context['version'] = VERSION.rsplit('.', 1)[0]
+                context['version'] = __version__.rsplit('.', 1)[0]
                 context['ref'] = self.mh.get_id
-                context['obj'] = lambda *a: 1
-                self.values[field_name] = safe_eval(eval_attr, context)
+                context['Decimal'] = Decimal
+                context['datetime'] = datetime
+                if pyson_attr:
+                    context.update(CONTEXT)
+                value = eval(eval_attr, context)
+                if pyson_attr:
+                    value = PYSONEncoder().encode(value)
+                self.values[field_name] = value
 
         else:
             raise Exception("Tags '%s' not supported inside tag record." %
@@ -235,7 +251,7 @@ class RecordTagHandler:
 
     def characters(self, data):
 
-        """If whe are in a field tag, consume all the content"""
+        """If we are in a field tag, consume all the content"""
 
         if not self.current_field:
             return
@@ -255,8 +271,7 @@ class RecordTagHandler:
 
         if name == "field":
             if not self.current_field:
-                raise Exception("Application error"
-                                "current_field expected to be set.")
+                return self
             # Escape end cdata tag :
             if self.cdata in ('inside', 'start'):
                 self.values[self.current_field] = \
@@ -284,8 +299,9 @@ class RecordTagHandler:
             raise Exception("Unexpected closing tag '%s'" % (name,))
 
     def current_state(self):
-        return "In tag record: model %s with id %s." % \
-               (self.model and self.model.__name__ or "?", self.xml_id)
+        return "In tag record model %s with id %s.%s." % \
+               (self.model and self.model.__name__ or "?",
+                   self.mh.module, self.xml_id)
 
 
 # Custom exception:
@@ -327,7 +343,7 @@ class Fs2bdAccessor:
         Whe call the prefetch function here to. Like that whe are sure
         not to erase data when get is called.
         """
-        if not module in self.fetched_modules:
+        if module not in self.fetched_modules:
             self.fetch_new_module(module)
         if fs_id not in self.fs2db[module]:
             self.fs2db[module][fs_id] = {}
@@ -341,20 +357,19 @@ class Fs2bdAccessor:
         self.browserecord[module].setdefault(model_name, {})
         Model = self.pool.get(model_name)
         if not ids:
-            ids = self.browserecord[module][model_name].keys()
+            ids = list(self.browserecord[module][model_name].keys())
         models = Model.browse(ids)
         for model in models:
             if model.id in self.browserecord[module][model_name]:
-                for cache in Transaction().cursor.cache.values():
-                    for cache in (cache, cache.get('_language_cache',
-                                {}).values()):
+                for cache in Transaction().cache.values():
+                    for cache in (
+                            cache, cache.get('_language_cache', {}).values()):
                         if (model_name in cache
                                 and model.id in cache[model_name]):
                             cache[model_name][model.id] = {}
             self.browserecord[module][model_name][model.id] = model
 
     def fetch_new_module(self, module):
-        cursor = Transaction().cursor
         self.fs2db[module] = {}
         module_data_ids = self.ModelData.search([
                 ('module', '=', module),
@@ -369,22 +384,20 @@ class Fs2bdAccessor:
             record_ids.setdefault(rec.model, [])
             record_ids[rec.model].append(rec.db_id)
 
-        object_name_list = set(self.pool.object_name_list())
-
         self.browserecord[module] = {}
         for model_name in record_ids.keys():
-            if model_name not in object_name_list:
+            try:
+                Model = self.pool.get(model_name)
+            except KeyError:
                 continue
-            Model = self.pool.get(model_name)
             self.browserecord[module][model_name] = {}
-            for i in range(0, len(record_ids[model_name]), cursor.IN_MAX):
-                sub_record_ids = record_ids[model_name][i:i + cursor.IN_MAX]
+            for sub_record_ids in grouped_slice(record_ids[model_name]):
                 with Transaction().set_context(active_test=False):
                     records = Model.search([
-                        ('id', 'in', sub_record_ids),
-                        ])
-                with Transaction().set_context(language='en_US'):
-                    models = Model.browse(map(int, records))
+                        ('id', 'in', list(sub_record_ids)),
+                        ], order=[('id', 'ASC')])
+                with Transaction().set_context(language='en'):
+                    models = Model.browse(list(map(int, records)))
                 for model in models:
                     self.browserecord[module][model_name][model.id] = model
         self.fetched_modules.append(module)
@@ -392,7 +405,7 @@ class Fs2bdAccessor:
 
 class TrytondXmlHandler(sax.handler.ContentHandler):
 
-    def __init__(self, pool, module, module_state):
+    def __init__(self, pool, module, module_state, modules):
         "Register known taghandlers, and managed tags."
         sax.handler.ContentHandler.__init__(self)
 
@@ -408,10 +421,7 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
         self.grouped_write = defaultdict(list)
         self.grouped_model_data = []
         self.skip_data = False
-        Module = pool.get('ir.module.module')
-        self.installed_modules = [m.name for m in Module.search([
-                    ('state', 'in', ['installed', 'to upgrade']),
-                    ])]
+        self.modules = modules
 
         # Tag handlders are used to delegate the processing
         self.taghandlerlist = {
@@ -440,8 +450,9 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
         try:
             self.sax_parser.parse(source)
         except Exception:
-            logging.getLogger("convert").error(
-                "Error while parsing xml file:\n" + self.current_state())
+            logger.error(
+                "Error while parsing xml file:\n" + self.current_state(),
+                exc_info=True)
             raise
         return self.to_delete
 
@@ -453,8 +464,6 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
 
             if name in self.taghandlerlist:
                 self.taghandler = self.taghandlerlist[name]
-                self.taghandler.startElement(name, attributes)
-
             elif name == "data":
                 self.noupdate = bool(int(attributes.get("noupdate", '0')))
                 self.grouped = bool(int(attributes.get('grouped', 0)))
@@ -464,19 +473,17 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
                 else:
                     self.skip_data = False
                 depends = attributes.get('depends', '').split(',')
-                depends = [m.strip() for m in depends if m]
-                if depends:
-                    if not all((m in self.installed_modules for m in depends)):
-                        self.skip_data = True
+                depends = {m.strip() for m in depends if m}
+                if not depends.issubset(self.modules):
+                    self.skip_data = True
 
             elif name == "tryton":
                 pass
 
             else:
-                logging.getLogger("convert").info("Tag %s not supported" %
-                    name)
+                logger.info("Tag %s not supported", (name,))
                 return
-        elif not self.skip_data:
+        if self.taghandler and not self.skip_data:
             self.taghandler.startElement(name, attributes)
 
     def characters(self, data):
@@ -486,14 +493,14 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
     def endElement(self, name):
 
         if name == 'data' and self.grouped:
-            for model, values in self.grouped_creations.iteritems():
+            for model, values in self.grouped_creations.items():
                 self.create_records(model, values.values(), values.keys())
             self.grouped_creations.clear()
-            for key, actions in self.grouped_write.iteritems():
+            for key, actions in self.grouped_write.items():
                 module, model = key
                 self.write_records(module, model, *actions)
             self.grouped_write.clear()
-        if self.grouped_model_data:
+        if name == 'data' and self.grouped_model_data:
             self.ModelData.write(*self.grouped_model_data)
             del self.grouped_model_data[:]
 
@@ -539,16 +546,7 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
         elif field_type == 'reference':
             if not getattr(record, key):
                 return None
-            elif not isinstance(getattr(record, key), basestring):
-                return str(getattr(record, key))
-            ref_mode, ref_id = getattr(record, key).split(',', 1)
-            try:
-                ref_id = safe_eval(ref_id)
-            except Exception:
-                pass
-            if isinstance(ref_id, (list, tuple)):
-                ref_id = ref_id[0]
-            return ref_mode + ',' + str(ref_id)
+            return str(getattr(record, key))
         elif field_type in ['one2many', 'many2many']:
             raise Unhandled_field("Unhandled field %s" % key)
         else:
@@ -592,7 +590,7 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
             if module == self.module and fs_id in self.to_delete:
                 self.to_delete.remove(fs_id)
 
-            if self.noupdate and self.module_state != 'to install':
+            if self.noupdate and self.module_state != 'to activate':
                 return
 
             # this record is already in the db:
@@ -604,13 +602,10 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
             if not old_values:
                 old_values = {}
             else:
-                old_values = safe_eval(old_values, {
-                    'Decimal': Decimal,
-                    'datetime': datetime,
-                    })
+                old_values = self.ModelData.load_values(old_values)
 
             for key in old_values:
-                if isinstance(old_values[key], str):
+                if isinstance(old_values[key], bytes):
                     # Fix for migration to unicode
                     old_values[key] = old_values[key].decode('utf-8')
 
@@ -620,10 +615,10 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
                     % (fs_id, module))
 
             record = self.fs2db.get_browserecord(module, Model.__name__, db_id)
-            #Re-create record if it was deleted
+            # Re-create record if it was deleted
             if not record:
                 with Transaction().set_context(
-                        module=module, language='en_US'):
+                        module=module, language='en'):
                     record, = Model.create([values])
 
                 # reset_browsercord
@@ -648,12 +643,8 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
 
                 db_field = self._clean_value(key, record)
 
-                # if the fs value is the same has in the db, whe ignore it
-                val = values[key]
-                if isinstance(values[key], str):
-                    # Fix for migration to unicode
-                    val = values[key].decode('utf-8')
-                if db_field == val:
+                # if the fs value is the same as in the db, we ignore it
+                if db_field == values[key]:
                     continue
 
                 # we cannot update a field if it was changed by a user...
@@ -663,22 +654,15 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
                 else:
                     expected_value = old_values[key]
 
-                # Migration from 2.0: Reference field change value
-                field_type = Model._fields[key]._type
-                if field_type == 'reference':
-                    if (expected_value and expected_value.endswith(',0')
-                            and not db_field):
-                        db_field = expected_value
-
                 # ... and we consider that there is an update if the
                 # expected value differs from the actual value, _and_
                 # if they are not false in a boolean context (ie None,
                 # False, {} or [])
                 if db_field != expected_value and (db_field or expected_value):
-                    logging.getLogger("convert").warning(
+                    logger.warning(
                         "Field %s of %s@%s not updated (id: %s), because "
-                        "it has changed since the last update"
-                        % (key, record.id, model, fs_id))
+                        "it has changed since the last update",
+                        key, record.id, model, fs_id)
                     continue
 
                 # so, the field in the fs and in the db are different,
@@ -687,10 +671,10 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
 
             if self.grouped:
                 self.grouped_write[(module, model)].extend(
-                    (record, to_update, old_values, fs_id, mdata_id))
+                    (record, to_update, old_values, values, fs_id, mdata_id))
             else:
-                self.write_records(module, model, record, to_update,
-                    old_values, fs_id, mdata_id)
+                self.write_records(module, model,
+                    record, to_update, old_values, values, fs_id, mdata_id)
         else:
             if self.grouped:
                 self.grouped_creations[model][fs_id] = values
@@ -700,11 +684,11 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
     def create_records(self, model, vlist, fs_ids):
         Model = self.pool.get(model)
 
-        with Transaction().set_context(module=self.module, language='en_US'):
+        with Transaction().set_context(module=self.module, language='en'):
             records = Model.create(vlist)
 
         mdata_values = []
-        for record, values, fs_id in izip(records, vlist, fs_ids):
+        for record, values, fs_id in zip(records, vlist, fs_ids):
             for key in values:
                 values[key] = self._clean_value(key, record)
 
@@ -713,38 +697,39 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
                     'model': model,
                     'module': self.module,
                     'db_id': record.id,
-                    'values': str(values),
+                    'values': self.ModelData.dump_values(values),
+                    'fs_values': self.ModelData.dump_values(values),
                     'noupdate': self.noupdate,
                     })
 
         models_data = self.ModelData.create(mdata_values)
 
-        for record, values, fs_id, mdata in izip(
+        for record, values, fs_id, mdata in zip(
                 records, vlist, fs_ids, models_data):
             self.fs2db.set(self.module, fs_id, {
                     'db_id': record.id,
                     'model': model,
                     'id': mdata.id,
-                    'values': str(values),
+                    'values': self.ModelData.dump_values(values),
                     })
         self.fs2db.reset_browsercord(self.module, model,
             [r.id for r in records])
 
     def write_records(self, module, model,
-            record, values, old_values, fs_id, mdata_id, *args):
-        args = (record, values, old_values, fs_id, mdata_id) + args
+            record, values, old_values, new_values, fs_id, mdata_id, *args):
+        args = (record, values, old_values, new_values, fs_id, mdata_id) + args
         Model = self.pool.get(model)
 
         actions = iter(args)
         to_update = []
-        for record, values, _, _, _ in zip(*((actions,) * 5)):
+        for record, values, _, _, _, _ in zip(*((actions,) * 6)):
             if values:
                 to_update += [[record], values]
         # if there is values to update:
         if to_update:
             # write the values in the db:
             with Transaction().set_context(
-                    module=module, language='en_US'):
+                    module=module, language='en'):
                 Model.write(*to_update)
             self.fs2db.reset_browsercord(
                 module, Model.__name__, sum(to_update[::2], []))
@@ -762,11 +747,13 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
                 values[key] = self._clean_value(key, record)
 
         actions = iter(args)
-        for record, values, old_values, fs_id, mdata_id in zip(
-                *((actions,) * 5)):
+        for record, values, old_values, new_values, fs_id, mdata_id in zip(
+                *((actions,) * 6)):
             temp_values = old_values.copy()
             temp_values.update(values)
             values = temp_values
+            fs_values = old_values.copy()
+            fs_values.update(new_values)
 
             if values != old_values:
                 self.grouped_model_data.extend(([self.ModelData(mdata_id)], {
@@ -774,19 +761,19 @@ class TrytondXmlHandler(sax.handler.ContentHandler):
                             'model': model,
                             'module': module,
                             'db_id': record.id,
-                            'values': str(values),
-                            'date_update': datetime.datetime.now(),
+                            'values': self.ModelData.dump_values(values),
+                            'fs_values': self.ModelData.dump_values(fs_values),
                             }))
 
         # reset_browsercord to keep cache memory low
-        self.fs2db.reset_browsercord(module, Model.__name__, args[::5])
+        self.fs2db.reset_browsercord(module, Model.__name__, args[::6])
 
 
 def post_import(pool, module, to_delete):
     """
     Remove the records that are given in to_delete.
     """
-    cursor = Transaction().cursor
+    transaction = Transaction()
     mdata_delete = []
     ModelData = pool.get("ir.model.data")
 
@@ -796,48 +783,47 @@ def post_import(pool, module, to_delete):
             ('module', '=', module),
             ], order=[('id', 'DESC')])
 
-    object_name_list = set(pool.object_name_list())
     for mrec in mdata:
         model, db_id, fs_id = mrec.model, mrec.db_id, mrec.fs_id
 
-        logging.getLogger("convert").info(
-            'Deleting %s@%s from %s.%s', db_id, model, module, fs_id)
+        logger.info('Deleting %s@%s from %s.%s', db_id, model, module, fs_id)
         try:
             # Deletion of the record
-            if model in object_name_list:
+            try:
                 Model = pool.get(model)
+            except KeyError:
+                Model = None
+            if Model:
                 Model.delete([Model(db_id)])
                 mdata_delete.append(mrec)
             else:
-                logging.getLogger("convert").warning(
-                        'Could not delete id %d of model %s because model no '
-                        'longer exists.' % (db_id, model))
-            cursor.commit()
+                logger.warning(
+                    'Could not delete id %d of model %s because model no '
+                    'longer exists.', db_id, model)
         except Exception:
-            cursor.rollback()
-            tb_s = ''.join(traceback.format_exception(*sys.exc_info()))
-            logging.getLogger("convert").error(
+            transaction.rollback()
+            logger.error(
                 'Could not delete id: %d of model %s\n'
-                    'There should be some relation '
-                    'that points to this resource\n'
-                    'You should manually fix this '
-                    'and restart --update=module\n'
-                    'Exception: %s' %
-                    (db_id, model, tb_s))
+                'There should be some relation '
+                'that points to this resource\n'
+                'You should manually fix this '
+                'and restart --update=module\n',
+                db_id, model, exc_info=True)
             if 'active' in Model._fields:
                 try:
                     Model.write([Model(db_id)], {
                             'active': False,
                             })
                 except Exception:
-                    cursor.rollback()
-                    logging.getLogger("convert").error(
+                    transaction.rollback()
+                    logger.error(
                         'Could not inactivate id: %d of model %s\n',
                         db_id, model, exc_info=True)
+        transaction.commit()
 
     # Clean model_data:
     if mdata_delete:
         ModelData.delete(mdata_delete)
-        cursor.commit()
+        transaction.commit()
 
     return True
