@@ -1,13 +1,13 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 import os
-import sys
 import logging
+import json
+
 from lxml import etree
 from trytond.model import ModelView, ModelSQL, fields
-from trytond import backend
-from trytond.pyson import CONTEXT, Eval, Bool, PYSONDecoder
-from trytond.tools import safe_eval, file_open
+from trytond.pyson import Eval, Bool, PYSONDecoder, If
+from trytond.tools import file_open
 from trytond.transaction import Transaction
 from trytond.wizard import Wizard, StateView, Button
 from trytond.pool import Pool
@@ -19,6 +19,8 @@ __all__ = [
     'ViewTreeWidth', 'ViewTreeState', 'ViewSearch',
     ]
 
+logger = logging.getLogger(__name__)
+
 
 class View(ModelSQL, ModelView):
     "View"
@@ -35,7 +37,13 @@ class View(ModelSQL, ModelView):
             ('graph', 'Graph'),
             ('calendar', 'Calendar'),
             ('board', 'Board'),
-            ], 'View Type', select=True)
+            ], 'View Type', select=True,
+        domain=[
+            If(Bool(Eval('inherit')),
+                ('type', '=', None),
+                ('type', '!=', None)),
+            ],
+        depends=['inherit'])
     data = fields.Text('Data')
     name = fields.Char('Name', states={
             'invisible': ~(Eval('module') & Eval('name')),
@@ -66,27 +74,10 @@ class View(ModelSQL, ModelView):
         cls._buttons.update({
                 'show': {
                     'readonly': Eval('type') != 'form',
+                    'depends': ['type'],
                     },
                 })
 
-    @classmethod
-    def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-        cursor = Transaction().cursor
-        table = TableHandler(cursor, cls, module_name)
-
-        # Migration from 2.4 arch moved into data
-        if table.column_exist('arch'):
-            table.column_rename('arch', 'data')
-
-        super(View, cls).__register__(module_name)
-
-        # Migration from 1.0 arch no more required
-        table.not_null_action('arch', action='remove')
-
-        # Migration from 2.4 model no more required
-        table.not_null_action('model', action='remove')
-
     @staticmethod
     def default_priority():
         return 16
@@ -105,13 +96,18 @@ class View(ModelSQL, ModelView):
         key = (cls.__name__, type_)
         rng = cls._get_rng_cache.get(key)
         if rng is None:
-            rng_name = os.path.join(os.path.dirname(
-                    unicode(__file__, sys.getfilesystemencoding())),
-                type_ + '.rng')
-            rng = etree.fromstring(open(rng_name).read())
+            rng_name = os.path.join(os.path.dirname(__file__), type_ + '.rng')
+            with open(rng_name, 'rb') as fp:
+                rng = etree.fromstring(fp.read())
             cls._get_rng_cache.set(key, rng)
         return rng
 
+    @property
+    def rng_type(self):
+        if self.inherit:
+            return self.inherit.rng_type
+        return self.type
+
     @classmethod
     def validate(cls, views):
         super(View, cls).validate(views)
@@ -129,15 +125,14 @@ class View(ModelSQL, ModelView):
             tree = etree.fromstring(xml)
 
             if hasattr(etree, 'RelaxNG'):
-                rng_type = view.inherit.type if view.inherit else view.type
-                validator = etree.RelaxNG(etree=cls.get_rng(rng_type))
+                validator = etree.RelaxNG(etree=cls.get_rng(view.rng_type))
                 if not validator.validate(tree):
-                    logger = logging.getLogger('ir')
-                    error_log = reduce(lambda x, y: str(x) + '\n' + str(y),
-                            validator.error_log.filter_from_errors())
-                    logger.error('Invalid xml view:\n%s'
-                        % (str(error_log) + '\n' + xml))
-                    cls.raise_user_error('invalid_xml', (view.rec_name,))
+                    error_log = '\n'.join(map(str,
+                            validator.error_log.filter_from_errors()))
+                    logger.error('Invalid XML view %s:\n%s\n%s',
+                        view.rec_name, error_log, xml)
+                    cls.raise_user_error(
+                        'invalid_xml', (view.rec_name,), error_log)
             root_element = tree.getroottree().getroot()
 
             # validate pyson attributes
@@ -146,31 +141,32 @@ class View(ModelSQL, ModelView):
             }
 
             def encode(element):
-                for attr in ('states', 'domain', 'spell', 'colors'):
-                    if element.get(attr):
-                        try:
-                            value = safe_eval(element.get(attr), CONTEXT)
-                            validates.get(attr, lambda a: True)(value)
-                        except Exception, e:
-                            logger = logging.getLogger('ir')
-                            logger.error('Invalid pyson view element "%s:%s":'
-                                '\n%s\n%s'
-                                % (element.get('id') or element.get('name'),
-                                    attr, str(e), xml))
-                            return False
+                for attr in ('states', 'domain', 'spell'):
+                    if not element.get(attr):
+                        continue
+                    try:
+                        value = PYSONDecoder().decode(element.get(attr))
+                        validates.get(attr, lambda a: True)(value)
+                    except Exception as e:
+                        error_log = '%s: <%s %s="%s"/>' % (
+                            e, element.get('id') or element.get('name'), attr,
+                            element.get(attr))
+                        logger.error(
+                            'Invalid XML view %s:\n%s\n%s',
+                            view.rec_name, error_log, xml)
+                        cls.raise_user_error(
+                            'invalid_xml', (view.rec_name,), error_log)
                 for child in element:
-                    if not encode(child):
-                        return False
-                return True
-            if not encode(root_element):
-                cls.raise_user_error('invalid_xml', (view.rec_name,))
+                    encode(child)
+            encode(root_element)
 
     def get_arch(self, name):
         value = None
         if self.name and self.module:
             path = os.path.join(self.module, 'view', self.name + '.xml')
             try:
-                with file_open(path, subdir='modules') as fp:
+                with file_open(path,
+                        subdir='modules', mode='r', encoding='utf-8') as fp:
                     value = fp.read()
             except IOError:
                 pass
@@ -216,10 +212,13 @@ class ShowView(Wizard):
         def __init__(self, model_name, buttons):
             StateView.__init__(self, model_name, None, buttons)
 
-        def get_view(self):
+        def get_view(self, wizard, state_name):
             pool = Pool()
             View = pool.get('ir.ui.view')
-            view = View(Transaction().context.get('active_id'))
+            view_id = Transaction().context.get('active_id')
+            if not view_id:
+                return {}
+            view = View(view_id)
             Model = pool.get(view.model)
             return Model.fields_view_get(view_id=view.id)
 
@@ -273,12 +272,12 @@ class ViewTreeWidth(ModelSQL, ModelView):
         records = cls.search([
             ('user', '=', Transaction().user),
             ('model', '=', model),
-            ('field', 'in', fields.keys()),
+            ('field', 'in', list(fields.keys())),
             ])
         cls.delete(records)
 
         to_create = []
-        for field in fields.keys():
+        for field in list(fields.keys()):
             to_create.append({
                     'model': model,
                     'field': field,
@@ -305,23 +304,15 @@ class ViewTreeState(ModelSQL, ModelView):
     def __setup__(cls):
         super(ViewTreeState, cls).__setup__()
         cls.__rpc__.update({
-                'set': RPC(readonly=False),
-                'get': RPC(),
+                'set': RPC(readonly=False, check_access=False),
+                'get': RPC(check_access=False),
                 })
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-        cursor = Transaction().cursor
-        table = TableHandler(cursor, cls, module_name)
-
-        # Migration from 2.8: table name changed
-        table.table_rename(cursor, 'ir_ui_view_tree_expanded_state',
-            cls._table)
-
         super(ViewTreeState, cls).__register__(module_name)
 
-        table = TableHandler(cursor, cls, module_name)
+        table = cls.__table_handler__(module_name)
         table.index_action(['model', 'domain', 'user', 'child_name'], 'add')
 
     @staticmethod
@@ -334,41 +325,43 @@ class ViewTreeState(ModelSQL, ModelView):
 
     @classmethod
     def set(cls, model, domain, child_name, nodes, selected_nodes):
+        # Normalize the json domain
+        domain = json.dumps(json.loads(domain), separators=(',', ':'))
         current_user = Transaction().user
-        with Transaction().set_user(0):
-            records = cls.search([
+        records = cls.search([
+                ('user', '=', current_user),
+                ('model', '=', model),
+                ('domain', '=', domain),
+                ('child_name', '=', child_name),
+                ])
+        cls.delete(records)
+        cls.create([{
+                    'user': current_user,
+                    'model': model,
+                    'domain': domain,
+                    'child_name': child_name,
+                    'nodes': nodes,
+                    'selected_nodes': selected_nodes,
+                    }])
+
+    @classmethod
+    def get(cls, model, domain, child_name):
+        # Normalize the json domain
+        domain = json.dumps(json.loads(domain), separators=(',', ':'))
+        current_user = Transaction().user
+        try:
+            expanded_info, = cls.search([
                     ('user', '=', current_user),
                     ('model', '=', model),
                     ('domain', '=', domain),
                     ('child_name', '=', child_name),
-                    ])
-            cls.delete(records)
-            cls.create([{
-                        'user': current_user,
-                        'model': model,
-                        'domain': domain,
-                        'child_name': child_name,
-                        'nodes': nodes,
-                        'selected_nodes': selected_nodes,
-                        }])
-
-    @classmethod
-    def get(cls, model, domain, child_name):
-        current_user = Transaction().user
-        with Transaction().set_user(0):
-            try:
-                expanded_info, = cls.search([
-                        ('user', '=', current_user),
-                        ('model', '=', model),
-                        ('domain', '=', domain),
-                        ('child_name', '=', child_name),
-                        ],
-                    limit=1)
-            except ValueError:
-                return (cls.default_nodes(), cls.default_selected_nodes())
-            state = cls(expanded_info)
-            return (state.nodes or cls.default_nodes(),
-                state.selected_nodes or cls.default_selected_nodes())
+                    ],
+                limit=1)
+        except ValueError:
+            return (cls.default_nodes(), cls.default_selected_nodes())
+        state = cls(expanded_info)
+        return (state.nodes or cls.default_nodes(),
+            state.selected_nodes or cls.default_selected_nodes())
 
 
 class ViewSearch(ModelSQL, ModelView):
